/*
**++
**  FACILITY:
**
**      Funktionsauswerter zum Definieren und Auswerten beliebiger 
**	reellwertiger Funktionen.
**	[@tbs@]...
**
**  ABSTRACT:
**
**      [@tbs@]...
**
**  AUTHORS:
**
**      M. Schmidt
**
**
**  CREATION DATE:      24.01.1990
**
**  MODIFICATION HISTORY:
**	 28.08.1990 Makro- und Funktionsdefinition und -Auswertung eingebaut
**	 23.07.1991 Definition und Auswertung tabellierter Funktionen eingebaut
**	 11.10.1991 Max und Min Funktionen ergaenzt
**
**--
*/


/*
**
**  INCLUDE FILES
**
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "fktdef.h"
#include "f_parse.h"

extern double atof();
double btof();

/*
**
**  MACRO DEFINITIONS
**
*/

#define MAXDIM	3
#define MAXFKT	5
#define MAXVARS	257

char *Formel = NULL, 
     *Fp = NULL, 
     *Ep = NULL, 
     *F_ep[MAXFKT],
     Param[MAXDIM];

Value  F_wert;

double F_max, 
       Parval[MAXDIM];

short F_init = 0, 
      F_lim  = 0, 
      F_ifk  = -1, 
      mfkpar = 0,
      F_dim  = 0;

int *S_tab,
    N_Fcall;

static Symbol *symlist = NULL;
static Psymbol *psymlist = NULL;

extern void	Ln(), Exp(), Sqrt(), Int(), Fak(), Hypot(),
		Sin(), Cos(), Tan(), Asin(), Acos(), Atan(), Atan2(), 
		Sinh(), Cosh(), Tanh(),
		Abs(), Max(), Min(),
		C_rnc(), C_cnr(), C_cni(), C_arg(), Konj(), Norm();

Const_1 consts_1[] = {
   { "pi",	3.14159265358979323846 },
   { "e",	2.71828182845904523536 },
   { "gamma",	0.57721566490153286060 },
   { "grad",	0.01745329251994 },
   { "gs",	0.6180339887499  },
   { "phi",	1.61803398874989484820 },
   NULL
};

static struct
{
   char *name;
   SKALAR cv[2];
} consts_2[] = {
   { "i",  0.0, 1.0 },
   NULL
};

static struct
{
   char *name;
   short narg;
   void (*func)();
} builtins[] = {
   { "abs", 1, Abs },
   { "int", 1, Int },
   { "fak", 1, Fak },
   { "sqrt", 1, Sqrt },
   { "hypot", MULTPAR, Hypot },
   { "max", MULTPAR, Max },
   { "min", MULTPAR, Min },
   { "ln", 1, Ln },
   { "exp", 1, Exp },
   { "sin", 1, Sin },
   { "cos", 1, Cos },
   { "tan", 1, Tan },
   { "asin", 1, Asin },
   { "acos", 1, Acos },
   { "atan", 1, Atan },
   { "atan2", 2, Atan2 },
   { "sinh", 1, Sinh },
   { "cosh", 1, Cosh },
   { "tanh", 1, Tanh },
   { "cx", 2, C_rnc },
   { "re", 1, C_cnr },
   { "im", 1, C_cni },
   { "arg", 1, C_arg },
   { "konj", 1, Konj },
   { "N", 1, Norm },
   NULL
};


F_ini( expr )
char *expr;
{
   int i, k;

   if (!F_init)
   {
      Symbol *s;
      extern int *table();

      S_tab = table( MAXVARS );

       for ( i=0; builtins[i].name; i++ )
       {
	  s = install( builtins[i].name, BLTIN, builtins[i].narg, 0 );
	  s->V.u.f = builtins[i].func;
       }

       for ( i=0; consts_1[i].name; i++ )
	   s = install( consts_1[i].name, CONST, 1, &consts_1[i].cval );

       for ( i=0; consts_2[i].name; i++ )
	   s = install( consts_2[i].name, CONST, 2, consts_2[i].cv );

      F_init = 1;
   }

   Ep = expr;
   k  = yyparse();
   Ep = NULL;
   F_ifk = -1;
   F_ep[0] = NULL;
   return k;
}

F_def (expr, vars, fmax)
char  *expr, *vars;
double fmax;
{
   short i;

   if (Formel)
      free(Formel);

   Formel = emalloc(strlen(expr)+1);
   Fp = strcpy(Formel,expr);

   for (F_dim = 0, i = 0;  vars[i] != EOS;  i++ )
      if (isalpha(vars[i]))
      {
	 Param[F_dim++] = vars[i];

	 while (isalpha(vars[i+1]))
	    i++;
      }

   if (fmax > .0)
   {
      F_max = fmax;
      F_lim = 1;
   }
   else
      F_lim = 0;

   N_Fcall = 0;
}

double F_val (p)
double p[];
{
   short i;

   for (i = 0;  i < F_dim;  i++ )
      Parval[i] = p[i];

   i = yyparse();
   N_Fcall++;

   if ( F_wert.type == 1 )
   {
      double w;

      w = F_wert.u.s;

      if (F_lim)
	 return ( w > F_max ? F_max : ( w < -F_max ? -F_max : w ));
      else
	 return w;
   }		     /* nicht skalare Werte z.Zt. nicht unterstuetzt ! */
   execerror( "Kein skalarer Wert!", (char *) 0 );
}

char F_input (void)
{
   if (F_ifk >= 0)
      if (F_ep[F_ifk])
	 return *(F_ep[F_ifk]++);

   if (Ep)
      return *(Ep++);

   if (Fp)
   {
      if (*Fp == EOS)
	 Fp = Formel;

      return *(Fp++);
   }

   return EOS;
}

void F_unput(c)
char c;
{
   if ( F_ifk >= 0)
      --F_ep[F_ifk];
   else if (Ep)
      --Ep;
   else if (Fp)
      --Fp;
}

Symbol *install( s, c, t, v )
char *s;
int c, t;
double *v;
{
   Symbol *sp;
   register int i;

   sp = (Symbol *) emalloc( sizeof( Symbol ));

   if ( tinsert( S_tab, s, sp ))
      execerror( "Fehler bei tinsert:", s );

   sp->name = emalloc( strlen(s)+1 );
   strcpy( sp->name, s );
   sp->class  = c;
   sp->V.type = t;

   if ( c == VAR || c == CONST )
      for ( i = 0; i < t; i++ )
	 sp->V.u.v[i] = v[i];

   sp->next = symlist;
   symlist = sp;
   return sp;
}

Psymbol *Pinstall(nam,val)	 /* aktuellen Parameter in temp. Tabelle eintragen */
char nam;
Value *val;
{
   short i;
   Psymbol *p;

   mfkpar++;
   p = (Psymbol *) emalloc(sizeof(Psymbol));
   p->pnam = nam;
   p->next = psymlist;
   psymlist = p;

   if ((p->pval.type = val->type) == 1 )
      p->pval.u.s = val->u.s;  
   else for ( i = 0; i < val->type; i++ )
      p->pval.u.v[i] = val->u.v[i];

   return p;
}

Psymbol * Plookup(pn)		 /* aktuellen Parameter in temp. Tabelle suchen */
char pn;
{
   Psymbol *ps;

   for ( ps = psymlist; ps; ps = ps->next )
      if ( ps->pnam == pn )
	 return ps;

   return NULL;
}

void Deinstall(pn)		 /* aktuellen Parameter aus temp. Tabelle entfernen */
char pn;
{
   Psymbol *ps, *pv;

   for ( pv = ps = psymlist; ps; ps = ps->next )
   {
      if ( ps->pnam == pn )
      {
	 if ( ps == psymlist )
	    psymlist = ps->next;
	 else
	    pv->next = ps->next;

	 free(ps);
	 --mfkpar;
	 break;
      }
      else
	 if ( ps != psymlist )
	    pv = pv->next;
   }
}

char *emalloc( n )
unsigned n;
{
   char *p, *malloc();

   p = malloc( n );

   if ( p == 0 )
      execerror( "Speichermangel", (char *) 0 );

   return p;
}

Psym(typ)
short typ;
{
   Symbol *sp;

   for ( sp = symlist; sp != (Symbol *) NULL; sp = sp->next )
      if ( sp->V.type == typ && ( sp->class == CONST || sp->class == VAR ))
	 switch (typ)
	 {
	 case 1:
	    printf( "\t%s:\t%.8g\n", sp->name, sp->V.u.s );
	    break;
	 case 2:
	    printf( "\t%s:\t(%.8g, %.8g)\n", sp->name, sp->V.u.v[0], sp->V.u.v[1] );
	    break;
	 }
      else if ( sp->class == MFKT && typ == 0 )
	 switch (sp->V.type)
	 {
	 case 0:
	    printf( "\t%s\t:: %s\n", sp->name, sp->V.u.mf.expr );
	    break;
	 case 1:
	    printf( "\t%s[%c]\t:: %s\n", sp->name, sp->V.u.mf.pnam[0], sp->V.u.mf.expr );
	    break;
	 case 2:
	    printf( "\t%s[%c,%c]\t:: %s\n", sp->name, sp->V.u.mf.pnam[0], sp->V.u.mf.pnam[1], sp->V.u.mf.expr );
	    break;
	 case 3:
	    printf( "\t%s[%c,%c,%c]\t:: %s\n", sp->name, sp->V.u.mf.pnam[0], sp->V.u.mf.pnam[1],
						sp->V.u.mf.pnam[2], sp->V.u.mf.expr );
	    break;
	 case 4:
	    printf( "\t%s[%c,%c,%c,%c]\t:: %s\n", sp->name, sp->V.u.mf.pnam[0], sp->V.u.mf.pnam[1], 
						sp->V.u.mf.pnam[2], sp->V.u.mf.pnam[3], sp->V.u.mf.expr );
	    break;
	 }

   printf("\n");
}

static char txtb[80], *tp = NULL;

double btof(fd)
FILE *fd;
{
   double w;

   if (tp)
      while (*tp <= ' ' && *tp > '\0')
         tp++;

   while (tp == NULL || strlen(tp) < 1)
   {
      if ((tp = fgets(txtb,80,fd)) == NULL)
      {
         printf("!! Lesefehler!\n");
	 exit();
      }

      while (*tp <= ' ' && *tp > '\0')
	 tp++;
   }

   w = atof(tp);
   tp = strchr(tp, ' ');
   return (w);
}

#define	k_ind(I,J,K) _d23*I + _d3*J + K	    /* Indexfunktion fuer Koeffizientenmatrizen */
#define	k_in2(I,J) _d2*I + J		    /* Indexfunktion fuer Koeffizientenmatrizen */

static short   _d2, _d3, _d23;	    /* Matrix-Kennwerte fuer Indexberechnung */
static short first=1;

void F_tdef( sf, dnam )
Symbol *sf;
char   *dnam;
{
   short i, j, k, it, dim[3];
   char	 ext[3];
   FILE	 *fd;

   if (( tp = strchr(dnam, '.')) == NULL)
   {
      printf("!! Fehler: Dateiname \"%s\" ungueltig!\n", dnam);
      return;
   }

   tp++;
   ext[0] = tolower(*tp++);
   ext[1] = tolower(*tp++);
   ext[2] = '\0';

   if (strncmp(ext, "mb", 2) == 0)
      sf->V.type = MB;
   else if (strncmp(ext, "bb", 2) == 0)
      sf->V.type = BB;
   else if (strncmp(ext, "mq", 2) == 0)
      sf->V.type = MQ;
   else
   {
      sf->V.type = 0;
      printf("!! Fehler: Dateityp \"%s\" ungueltig!\n", ext);
      return;
   }

   if (fd = fopen(dnam,"r"))
   {
      TabMQ    *mq;
      TabFP    *fp;

      tp = NULL;

      switch (sf->V.type)
      {
         case MB:
         case BB:
            sf->V.u.tfp = fp = (TabFP *) emalloc( sizeof( TabFP ));

            for (it = 1, i = 0;  i < 3;  i++)
            {
               fp->mg[i] = dim[i] = (short) btof(fd);
	       it *= (dim[i] + 1);
            }

	    _d2 = dim[1];  _d3 = dim[2];  _d23 = _d2 * _d3;
	    fp->km = (SKALAR *) emalloc( it * sizeof( SKALAR ));

            for (i = 0;  i < dim[0];  i++)
	       for (j = 0;  j < dim[1];  j++)
		  for (k = 0;  k < dim[2];  k++)
		     fp->km[k_ind(i,j,k)] = btof(fd);

            if (sf->V.type == BB)
            {
	       it = (dim[0] + 1) * (dim[1] + 1);
	       fp->b2 = (SKALAR *) emalloc( it * sizeof( SKALAR ));
	       it = (dim[0] + 1);
	       fp->b1 = (SKALAR *) emalloc( it * sizeof( SKALAR ));
	       it = (dim[0] > dim[1] ? (dim[0] > dim[2] ? dim[0] : dim[2])
				     : (dim[1] > dim[2] ? dim[1] : dim[2])) + 1;
	       fp->b0 = (SKALAR *) emalloc( it * sizeof( SKALAR ));
            }
	    else
            {
	       short *am2, *am1, s1, s2;

	       it = (dim[0] + 1) * (dim[1] + 1);
	       fp->b2 = (SKALAR *) emalloc( it * sizeof( short ));
	       am2 = (short *) fp->b2;
	       it = (dim[0] + 1);
	       fp->b1 = (SKALAR *) emalloc( it * sizeof( short ));
	       am1 = (short *) fp->b1;

	       for (i = 0;  i < dim[0];  i++)
	       {
		  for (s1 = j = 0;  j < dim[1];  j++)
		  {
		     for (s2 = k = 0;  k < dim[2] && !s2;  k++)
			s2 |= (fabs(fp->km[k_ind(i,j,k)]) > EPSD);

		     s1 |= am2[k_in2(i,j)] = s2;
		  }

		  am1[i] = s1;
	       }
            }

	    first = 1;
	    break;

         case MQ:
            sf->V.u.tmq = mq = (TabMQ *) emalloc( sizeof( TabMQ ));
	    mq->rad = btof(fd);

	    if ((mq->tlg = (int) btof(fd)) > MAXTEIL)
	    {
	       printf("!! Fehler: Zu viele Abschnitte!\n");
	       exit();
	    }

	    for (it = 0;  it < mq->tlg;  it++)
	    {
	       mq->zmin[it] = btof(fd);
	       mq->zmax[it] = btof(fd);
	       mq->scal[it] = btof(fd);

	       if ((mq->ianz[it] = (int) btof(fd)) > MAXPKT)
		  printf("!! Fehler: Zu viele Punkte in Abschnitt %d!\n", it+1);

	       for (i = 0;  i < mq->ianz[it];  i++)
	       {
		  mq->p[it][i][0] = btof(fd);
		  mq->p[it][i][1] = btof(fd);
		  mq->p[it][i][2] = btof(fd);
		  mq->a[it][i] = btof(fd);
		  mq->r[it][i] = btof(fd);
	       }
	    }
      }

      fclose(fd);
      sf->class = TABF;
   }
   else
   {
      printf("!! Fehler: Datei \"%s\" nicht zu oeffnen!\n", dnam);
      exit();
   }
}

static SKALAR *b0;   /* Verweis auf Arbeitsspeicher (vom Rufer bereitgestellt!) */

SKALAR casteljau( b, n, v0, v1 )
SKALAR b[], v0, v1;
short  n;
{
   short i, j;

   if ( n == 0 )
      return ( b[0] );

   for (i = 0;  i < n;  i++)
      b0[i] = v1 * b[i] + v0 * b[i+1];

   for (j = n - 1;  j > 0;  j--)
      for (i = 0;  i < j;  i++)
	 b0[i] = v1 * b0[i] + v0 * b0[i+1];
   
   return ( b0[0] );
}

void F_tval(r, sf, x, y, z)
Value *r, *x, *y, *z;
Symbol *sf;
{
   short    i, j, k;
   SKALAR   u, v, w;

   if (x->type == 1 && x->type == y->type && x->type == z->type)
   {
      SKALAR   p1, p2, p3, *a, *b, *b1, *b2, hf[2], zz;
      short    d1, d2, d3, ir, it, tre[2], *am1, *am2;
      TabMQ    *mq;

      p1 = x->u.s;
      p2 = y->u.s;
      p3 = z->u.s;

      switch (sf->V.type)
      {
         case MB:
            a  = sf->V.u.tfp->km;
            am1 = (short *) sf->V.u.tfp->b1;
            am2 = (short *) sf->V.u.tfp->b2;
            d1 = sf->V.u.tfp->mg[0];
            d2 = sf->V.u.tfp->mg[1];
            d3 = sf->V.u.tfp->mg[2];
	    _d2 = d2;  _d3 = d3;  _d23 = _d2 * _d3;

	    if (first)
	    {
	       printf("Koeffizientenmatrix (%d x %d x %d, Monombasis):\n",
			d1, d2, d3);

	       for (i = 0;  i < d1;  i++)
	       {
		  printf("am1(%d) = %d ", i, am1[i]);

		  for (j = 0;  j < d2;  j++)
		  {
		     printf("am2(%d,%d) = %d\n", i, j, am2[k_in2(i,j)]);

		     for (k = 0;  k < d3;  k++)
			printf("%g ", a[k_ind(i,j,k)]);

		     printf("\n");
		  }
	       }
	       first = 0;
	    }

	    if ( fabs(p1) < EPSD)
	       d1 = 1;

	    if ( fabs(p2) < EPSD)
	       d2 = 1;

	    if ( fabs(p3) < EPSD)
	       d3 = 1;

            for (u = 0.0, i = d1 - 1;  i >= 0;  i--)   /* Auswertung mittels Horner-Verfahren */
            {
	       v = 0.0;

	       if (am1[i])		     /* min. ein a(i,.,.) ~= 0 ? */
		  for (j = d2 - 1;  j >= 0;  j--)
		  {
		     w = 0.0;

		     if (am2[k_in2(i,j)])    /* min. ein a(i,j,.) ~= 0 ? */
			for (k = d3 - 1;  k >= 0;  k--)
			   w = w * p3 + a[k_ind(i,j,k)];

		     v = v * p2 + w;
		  }

	       u = u * p1 + v;
            }

	    r->u.s = u;
	    break;

         case BB:
            b  = sf->V.u.tfp->km;
            b0 = sf->V.u.tfp->b0;
            b1 = sf->V.u.tfp->b1;
            b2 = sf->V.u.tfp->b2;
            d1 = sf->V.u.tfp->mg[0];
            d2 = sf->V.u.tfp->mg[1];
            d3 = sf->V.u.tfp->mg[2];
	    _d2 = d2;  _d3 = d3;  _d23 = _d2 * _d3;
	    u  = p3; v = 1.0 - u;

	    if (first)
	    {
	       printf("Koeffizientenmatrix (%d x %d x %d, Bernsteinbasis):\n",
			    d1, d2, d3);

	       for (i = 0;  i < d1;  i++)
		  for (j = 0;  j < d2;  j++)
		  {
		     for (k = 0;  k < d3;  k++)
			printf("%g ", b[k_ind(i,j,k)]);

		     printf("\n");
		  }
	       first = 0;
	    }

            if (fabs(u) < EPSD)
               k = 0;
            else if (fabs(v) < EPSD)
               k = d3-1;
            else
               k = -1;

	    if (k < 0)
	       for (i = 0;  i < d1;  i++)
		  for (j = 0;  j < d2;  j++)
		     b2[k_in2(i,j)] = casteljau( &b[k_ind(i,j,0)], d3-1, u, v);
	    else
	       for (i = 0;  i < d1;  i++)
		  for (j = 0;  j < d2;  j++)
		     b2[k_in2(i,j)] = b[k_ind(i,j,k)];

	    u  = p2; v = 1.0 - u;

            if (fabs(u) < EPSD)
               k = 0;
            else if (fabs(v) < EPSD)
               k = d2-1;
            else
               k = -1;

	    if (k < 0)
	       for (i = 0;  i < d1;  i++)
		  b1[i] = casteljau( &b2[k_in2(i,0)], d2-1, u, v);
	    else
	       for (i = 0;  i < d1;  i++)
		  b1[i] = b2[k_in2(i,k)];

	    u  = p1; v = 1.0 - u;

            if (fabs(u) < EPSD)
               k = 0;
            else if (fabs(v) < EPSD)
               k = d1-1;
            else
               k = -1;

	    if (k < 0)
	       r->u.s = casteljau(b1, d1-1, u, v);
	    else
	       r->u.s = b1[k];

	    break;

         case MQ:
	    mq = sf->V.u.tmq;

	    for (ir = it = 0; it < mq->tlg;  it++ )
	       if ( mq->zmin[it] <= p3 && p3 <= mq->zmax[it])
		  tre[ir++] = it;

	    for (k = 0;  k < ir;  k++)
	    {
	       hf[k] = 0.0;
	       j = tre[k];

	       for (i = 0;  i < mq->ianz[j];  i++)
		  hf[k] += (sqrt( pow(p1 - mq->p[j][i][0], 2.0) +
				   pow(p2 - mq->p[j][i][1], 2.0) +
				  pow((p3 - mq->p[j][i][2]) * mq->scal[j], 2.0) +
				  pow(mq->r[j][i], 2.0)) * mq->a[j][i]);
	    }


	    if ( ir == 0 )
	       r->u.s = 0.0;	      /* Funktionswert hier nicht definiert ! */
	    else if (ir == 1)
	       r->u.s = hf[0];
	    else
	    {
	       zz = (p3 - mq->zmin[tre[1]]) / (mq->zmax[tre[0]] - mq->zmin[tre[1]]);
	       r->u.s = ((1.0 - 3.0 * pow(zz, 2.0) + 2.0 * pow(zz, 3.0)) * hf[0] +
			(	3.0 * pow(zz, 2.0) - 2.0 * pow(zz, 3.0)) * hf[1]);
	    }
	    break;
      }
   }
   else
      execerror("F_tvar", "unzulaessige(s) Argument(e)");

   r->type = x->type;
}
