#include <math.h>
#include <errno.h>
#ifdef FKT
#include "fktdef.h"
#else
#include "hoc.h"
#endif
#include "math_a.h"

/* extern int errno; */

#define PI   consts_1[0].cval
#define GRAD consts_1[3].cval

extern Const_1 consts_1[];
extern short   w_in_grad;

SKALAR errcheck(d,s)
SKALAR d;
char *s;
{
   if ( errno == EDOM )
   {
      errno = 0;
      execerror( s, "Argument nicht im Definitionsbereich" );
   }
   else if ( errno == ERANGE )
   {
      errno = 0;
      execerror( s, "Ergebnis undefiniert" );
   }

   return d;
}

double gnr(w)
double w;
{
   return ( w_in_grad ? w * GRAD : w );
}

double rng(w)
double w;
{
   return ( w_in_grad ? w / GRAD : w );
}

void Abs(r,x)
Value *r, *x;
{
   switch ( x->type )
   {
      case 1: r->u.s = fabs(x->u.s);
	      break;
      case 2: r->u.s = Cabs(x->u.v);
	      break;
      case 3: r->u.s = Vabs(x->u.v);
	      break;
      default: execerror( "abs", "unzulaessiger Argumenttyp" );
   }

   r->type = 1;
}

void Ln(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = errcheck( log(x->u.s), "ln" );
   else if ( x->type == 2 )
   {
      r->u.v[0] = errcheck( log(Cabs(x->u.v)), "ln" );
      r->u.v[1] = Carg(x->u.v);
   }
   else
      execerror( "ln", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Exp(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = errcheck( exp(x->u.s), "exp" );
   else if ( x->type == 2 )
   {
      double a;

      a = errcheck( exp(x->u.v[0]), "exp" );
      r->u.v[0] = a * cos(x->u.v[1]);
      r->u.v[1] = a * sin(x->u.v[1]);
   }
   else
      execerror( "exp", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Sqrt(r,x)
Value *r, *x;
{
   short t;

   if (( t = x->type ) == 1 )
   {
      if ( x->u.s < .0 )
      {
	 r->u.v[0] = 0;
	 r->u.v[1] = errcheck( sqrt(fabs(x->u.s)), "sqrt" );
	 t = 2;
      }
      else
	 r->u.s = errcheck( sqrt(x->u.s), "sqrt" );
   }
   else if ( x->type == 2 )
   {
      double a,b;

      a = errcheck( sqrt(Cabs(x->u.v)), "sqrt" );
      b = Carg(x->u.v) * 0.5;
      r->u.v[0] = a * cos(b);
      r->u.v[1] = a * sin(b);
   }
   else
      execerror( "sqrt", "unzulaessiger Argumenttyp" );

   r->type = t;
}

void Pow(r,x,y)
Value *r, *x, *y;
{
   if ( x->type == 1 && x->type == y->type )
      if ( x->u.s < .0 && y->u.s < 1.0 )
      {
	 x->u.v[1] = .0;
	 x->type   = 2;
      }

   if ( x->type == 1 && x->type == y->type )
      r->u.s = errcheck( pow(x->u.s,y->u.s), "exponentiation" );
   else if ( x->type == 2 )
   {
      double a, b;

      if ( y->type == 1 )
      {
	 b = Cabs(x->u.v);
	 b = errcheck( pow(b,y->u.s), "exponentiation" );
	 a = Carg(x->u.v) * y->u.s;
	 r->u.v[0] = b * cos(a);
	 r->u.v[1] = b * sin(a);
      }
      else if ( y->type == 2 )
      {
	 SKALAR h[2];

	 h[0] = log(Cabs(x->u.v));
	 h[1] = Carg(x->u.v);
	 Cmul(x->u.v,h,r->u.v);
	 b = errcheck( exp(r->u.v[0]), "exponentiation" );
	 a = r->u.v[1];
	 r->u.v[0] = b * cos(a);
	 r->u.v[1] = b * sin(a);
      }
      else
	 execerror( "Exponentiation", "unzulaessige Operation" );
   }
   else if ( x->type == 5 && y->type == 1 )
      Qpow(x->u.v,y->u.s,r->u.v);
   else
      execerror( "Exponentiation", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Int(r,x)
Value *r, *x;
{
   if ( x->type != 1 )
      execerror( "int", "unzulaessiger Argumenttyp" );

   r->u.s = (double) (long) x->u.s;
   r->type = 1;
}

#define MAX(a,b)    (a < b ? b : a)
#define MIN(a,b)    (a < b ? a : b)

void Max(r,x,y,z,w)
Value *r, *x, *y, *z, *w;
{
   switch ( r->type )
   {
      case 2:  if ( x->type == 1 && x->type == y->type )
		  r->u.s = MAX( x->u.s, y->u.s );
	       else
		  execerror( "max", "unzulaessiger Argumenttyp" );

	       break;
      case 3:  if ( x->type == 1 && x->type == y->type && y->type == z->type )
		  r->u.s = MAX( x->u.s, MAX( y->u.s, z->u.s ));
	       else
		  execerror( "max", "unzulaessiger Argumenttyp" );

	       break;
      case 4:  if ( x->type == 1 && x->type == y->type && y->type == z->type &&
		    z->type == w->type )
		  r->u.s = MAX( MAX( x->u.s, y->u.s), MAX( z->u.s, w->u.s ));
	       else
		  execerror( "max", "unzulaessiger Argumenttyp" );

	       break;
      default: execerror( "max", "unzulaessige Parameteranzahl" );
   }

   r->type = x->type;
}

void Min(r,x,y,z,w)
Value *r, *x, *y, *z, *w;
{
   switch ( r->type )
   {
      case 2:  if ( x->type == 1 && x->type == y->type )
		  r->u.s = MIN( x->u.s, y->u.s );
	       else
		  execerror( "min", "unzulaessiger Argumenttyp" );

	       break;
      case 3:  if ( x->type == 1 && x->type == y->type && y->type == z->type )
		  r->u.s = MIN( x->u.s, MIN( y->u.s, z->u.s ));
	       else
		  execerror( "min", "unzulaessiger Argumenttyp" );

	       break;
      case 4:  if ( x->type == 1 && x->type == y->type && y->type == z->type &&
		    z->type == w->type )
		  r->u.s = MIN( MIN( x->u.s, y->u.s), MIN( z->u.s, w->u.s ));
	       else
		  execerror( "min", "unzulaessiger Argumenttyp" );

	       break;
      default: execerror( "min", "unzulaessige Parameteranzahl" );
   }

   r->type = x->type;
}

void Fak(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
   {
      double F;
      int    ix;

      if ( x->u.s <= MAXFAKULT )
      {
	 for ( F = 1.0, ix = 2; ix <= x->u.s; ix++ )
	    F *= ix;

	 r->u.s = F;
      }
      else
	 execerror( "fak", "Argument zu gross" );
   }
   else
      execerror( "fak", "unzulaessiger Argumenttyp" );

   r->type = 1;
}

void Sin(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = sin(gnr(x->u.s));
   else if ( x->type == 2 )
   {
      r->u.v[0] = sin(x->u.v[0]) * cosh(x->u.v[1]);
      r->u.v[1] = cos(x->u.v[0]) * sinh(x->u.v[1]);
   }
   else
      execerror( "sin", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Cos(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = cos(gnr(x->u.s));
   else if ( x->type == 2 )
   {
      r->u.v[0] =  cos(x->u.v[0]) * cosh(x->u.v[1]);
      r->u.v[1] = -sin(x->u.v[0]) * sinh(x->u.v[1]);
   }
   else
      execerror( "cos", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Tan(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = errcheck( tan(gnr(x->u.s)), "tan" );
   else if ( x->type == 2 )
   {
      double a, x2, y2;

      x2 = 2.0 * x->u.v[0];
      y2 = 2.0 * x->u.v[1];
      a  = 1.0 / ( cos(x2) + cosh(y2));
      r->u.v[0] = a * sin(x2);
      r->u.v[1] = a * sinh(y2);
   }
   else
      execerror( "tan", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Sinh(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = sinh(gnr(x->u.s));
   else if ( x->type == 2 )
   {
      r->u.v[0] = sinh(x->u.v[0]) * cos(x->u.v[1]);
      r->u.v[1] = cosh(x->u.v[0]) * sin(x->u.v[1]);
   }
   else
      execerror( "sinh", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Cosh(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = cosh(gnr(x->u.s));
   else if ( x->type == 2 )
   {
      r->u.v[0] =  cosh(x->u.v[0]) * cos(x->u.v[1]);
      r->u.v[1] =  sinh(x->u.v[0]) * sin(x->u.v[1]);
   }
   else
      execerror( "cosh", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Tanh(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = errcheck( tanh(gnr(x->u.s)), "tanh" );
   else if ( x->type == 2 )
   {
      double a, x2, y2;

      x2 = 2.0 * x->u.v[0];
      y2 = 2.0 * x->u.v[1];
      a  = 1.0 / ( cosh(x2) + cos(y2));
      r->u.v[0] = a * sinh(x2);
      r->u.v[1] = a * sin(y2);
   }
   else
      execerror( "tanh", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Asin(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = rng(errcheck( asin(x->u.s), "asin" ));
   else
      execerror( "asin", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Acos(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = rng(errcheck( acos(x->u.s), "acos" ));
   else
      execerror( "acos", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Atan(r,x)
Value *r, *x;
{
   if ( x->type == 1 )
      r->u.s = rng(errcheck( atan(x->u.s), "atan" ));
   else
      execerror( "atan", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Atan2(r,x,y)
Value *r, *x, *y;
{
   if ( x->type == 1 && x->type == y->type )
      r->u.s = rng(errcheck( atan2(x->u.s,y->u.s), "atan2" ));
   else
      execerror( "atan2", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Hypot(r,x,y,z)
Value *r, *x, *y, *z;
{
   switch ( r->type )
   {
      case 2:  if ( x->type == 1 && x->type == y->type )
		  r->u.s = hypot(x->u.s,y->u.s);
	       else
		  execerror( "hypot", "unzulaessiger Argumenttyp" );

	       break;
      case 3:  if ( x->type == 1 && x->type == y->type && y->type == z->type )
		  r->u.s = sqrt(pow(x->u.s,2)+pow(y->u.s,2)+pow(z->u.s,2));
	       else
		  execerror( "hypot", "unzulaessiger Argumenttyp" );

	       break;
      default: execerror( "hypot", "unzulaessige Parameteranzahl" );
   }

   r->type = x->type;
}

void C_rnc(r,x,y)
Value *r, *x, *y;
{
   if ( x->type == 1 && x->type == y->type )
   {
      r->u.v[0] = x->u.s;
      r->u.v[1] = y->u.s;
   }
   else
      execerror( "cx", "unzulaessige Argumenttypen" );

   r->type = 2;
}

void C_cnr(r,x)
Value *r, *x;
{
   if ( x->type == 2 )
      r->u.s = x->u.v[0];
   else
      execerror( "re", "unzulaessiger Argumenttyp" );

   r->type = 1;
}

void C_cni(r,x)
Value *r, *x;
{
   if ( x->type == 2 )
      r->u.s = x->u.v[1];
   else
      execerror( "im", "unzulaessiger Argumenttyp" );

   r->type = 1;
}

void C_arg(r,x)
Value *r, *x;
{
   if ( x->type == 2 )
      r->u.s = rng(Carg(x->u.v));
   else
      execerror( "arg", "unzulaessiger Argumenttyp" );

   r->type = 1;
}

void Konj(r,x)
Value *r, *x;
{
   if ( x->type == 2 )
   {
      r->u.v[0] = x->u.v[0];
      r->u.v[1] = -x->u.v[1];
   }
   else if ( x->type == 5 )
   {
      short i;

      r->u.v[0] = x->u.v[0];

      for ( i = 1; i < 4; i++ )
         r->u.v[i] = -x->u.v[i];
   }
   else
      execerror( "konj", "unzulaessiger Argumenttyp" );

   r->type = x->type;
}

void Komp(i,f,r,x)
short i;    /* Komponenten-Art: 0: x, 1: y, 2: z, 3: u, 4: v, 5: r, 6: q */
char *f;
Value *r, *x;
{
   if ( x->type != 3 )
      execerror( f, "unzulaessiger Argumenttyp" );

   switch ( i )
   {
      double u[3];

   case 0:
   case 1:
   case 2:
      r->u.s = x->u.v[i];
      break;
   case 3:
   case 4:  
   case 5:
      Vkaku(x->u.v, &u[0], &u[1], &u[2] );
      r->u.s = ( i == 5 ? u[i-3] : rng(u[i-3]));
      break;
   case 6:
      r->u.s = hypot( x->u.v[0], x->u.v[1] );
      break;
   }

   r->type = 1;
}

void X_komp(r,x)
Value *r,*x;
{
   Komp(0,"X",r,x);
}

void Y_komp(r,x)
Value *r,*x;
{
   Komp(1,"Y",r,x);
}

void Z_komp(r,x)
Value *r,*x;
{
   Komp(2,"Z",r,x);
}

void U_komp(r,x)
Value *r,*x;
{
   Komp(3,"U",r,x);
}

void V_komp(r,x)
Value *r,*x;
{
   Komp(4,"V",r,x);
}

void R_komp(r,x)
Value *r,*x;
{
   Komp(5,"R",r,x);
}

void Q_komp(r,x)
Value *r,*x;
{
   Komp(6,"Q",r,x);
}

void Norm(r,x)
Value *r,*x;
{
   short i;

   for ( i = 0; i <  x->type; i++ )
      if ( fabs(x->u.v[i]) <= EPSD )
	 r->u.v[i] = 0.0;
      else
	 r->u.v[i] = x->u.v[i];

   r->type = x->type;
}

void T_punkt(r,x)
Value *r, *x;
{
   short i;

   if ( x->type == 4 )
      for ( i = 0; i < 3; i++ )
	 r->u.v[i] = x->u.v[i] * x->u.v[3];
   else if ( x->type == 6 )
      for ( i = 0; i < 3; i++ )
	 r->u.v[i] = x->u.v[i+3];
   else
      execerror( "tp", "Operation unzulaessig" );   

   r->type = 3;
}

void V_kart(r,x,y,z)
Value *r,*x,*y,*z;
{
   if ( x->type == 1 && x->type == y->type && y->type == z->type )
   {
      r->u.v[0] = x->u.s;
      r->u.v[1] = y->u.s;
      r->u.v[2] = z->u.s;
   }
   else
      execerror( "P", "unzulaessiger Argumenttyp" );

   r->type = 3;
}

void V_sph(r,x,y,z)
Value *r,*x,*y,*z;
{
   if ( x->type == 1 && x->type == y->type && y->type == z->type )
   {
      Vkuka(gnr(x->u.s),gnr(y->u.s),z->u.s,r->u.v);
   }
   else
      execerror( "P@", "unzulaessiger Argumenttyp" );

   r->type = 3;
}

void E_kart(r,x,y)
Value *r,*x,*y;
{
   if ( x->type == 3 && x->type == y->type )
   {
      Vricht(y->u.v,r->u.v);
      r->u.v[3] = Vskal(x->u.v,r->u.v);
   }
   else
      execerror( "E", "unzulaessiger Argumenttyp" );

   r->type = 4;
}

void E_sph(r,x,y,z)
Value *r,*x,*y,*z;
{
   if ( x->type == 3 && y->type == 1 && y->type == z->type )
   {
      Vkuka(gnr(y->u.s),gnr(z->u.s),1.0,r->u.v);
      r->u.v[3] = Vskal(x->u.v,r->u.v);
   }
   else
      execerror( "E@", "unzulaessiger Argumenttyp" );

   r->type = 4;
}

void Q_kart(r,x,y,z)
Value *r,*x,*y,*z;
{
   if ( r->type == 2 )
      if ( x->type == 3 && y->type == 1 )
      {
	 short  i;
	 double sw;

	 Vricht(x->u.v,&r->u.v[1]);
	 sw = gnr(y->u.s*0.5);
	 r->u.v[0] = cos(sw);
	 sw = sin(sw);

	 for ( i = 1; i < 4; i++ )
	    r->u.v[i] *= sw;
      }
      else
	 execerror( "rot", "unzulaessiger Argumenttyp" );
   else if ( r->type == 3 )
      if ( x->type == 3 && x->type == y->type && y->type == z->type )
      {
	 double sp;

	 sp = Vspat(x->u.v,y->u.v,z->u.v);

	 if ( fabs( sp - 1.0 ) < 0.0001 )
	 {
	    double A[3][3];
	    short  i, j;

	    for ( i = 0; i < 3; i++ )
	    {
	       A[i][0] = x->u.v[i];
	       A[i][1] = y->u.v[i];
	       A[i][2] = z->u.v[i];
	    }

	    Qmnq(A,r->u.v);
	 }
	 else
	    execerror( "rot", "Argumente muessen Spalten einer Drehmatrix sein" );
      }
      else
	 execerror( "rot", "unzulaessiger Argumenttyp" );
   else
      execerror( "rot", "unzulaessige Parameteranzahl" );

   r->type = 5;
}

void Q_sph(r,x,y,z)
Value *r,*x,*y,*z;
{
   if ( x->type == 1 && x->type == y->type && y->type == z->type )
   {
      short  i;
      double sw;

      Vkuka(gnr(x->u.s),gnr(y->u.s),1.0,&r->u.v[1]);
      sw = gnr(z->u.s*0.5);
      r->u.v[0] = cos(sw);
      sw = sin(sw);

      for ( i = 1; i < 4; i++ )
         r->u.v[i] *= sw;
   }
   else
      execerror( "rot@", "unzulaessiger Argumenttyp" );

   r->type = 5;
}

void G_norm(g)
Value *g;
{
   SKALAR a;
   short  i;

   Vricht(g->u.v,g->u.v);
   a = Vskal(&g->u.v[3],g->u.v);

   for ( i = 0; i < 3; i++ )
      g->u.v[i+3] -= a * g->u.v[i];
}

void G_kart(r,x,y)
Value *r,*x,*y;
{
   if ( x->type == 3 && x->type == y->type )
   {
      short i;

      for ( i = 0; i < 3; i++ )
      {
         r->u.v[i]   = x->u.v[i];
         r->u.v[i+3] = y->u.v[i];
      }
   }
   else
      execerror( "G", "unzulaessiger Argumenttyp" );

   r->type = 6;
   G_norm(r);
}

void G_sph(r,x,y,z)
Value *r,*x,*y,*z;
{
   if ( x->type == 3 && y->type == 1 && y->type == z->type )
   {
      short i;

      Vkuka(gnr(y->u.s),gnr(z->u.s),1.0,r->u.v);

      for ( i = 0; i < 3; i++ )
         r->u.v[i+3] = x->u.v[i];
   }
   else
      execerror( "G@", "unzulaessiger Argumenttyp" );

   r->type = 6;
   G_norm(r);
}

void Dist(r,x,y)
Value *r, *x, *y;
{
   SKALAR d;

   switch ( x->type )
   {
   case 3:
      switch ( y->type )
      {
      case 3:
	 d = Vdist( x->u.v, y->u.v );		   break;
      case 4:
	 d = Vdpe( x->u.v, y->u.v );		   break;
      case 6:
	 d = Vdpg( x->u.v, &y->u.v[3], y->u.v );   break;
      default:
	 execerror("D", "unzulaessige(s) Argument(e)" );
      }
      break;

   case 4:
      switch ( y->type )
      {
      case 3:
	 d = Vdpe( y->u.v, x->u.v );		   break;
      case 4:
	 d = Vdee( x->u.v, y->u.v );		   break;
      case 6:
	 d = Vdge( &y->u.v[3], y->u.v, x->u.v );   break;
      default:
	 execerror("D", "unzulaessige(s) Argument(e)" );
      }
      break;

   case 6:
      switch ( y->type )
      {
      case 3:
	 d = Vdpg( y->u.v, &x->u.v[3], x->u.v );   break;
      case 4:
	 d = Vdge( &x->u.v[3], x->u.v, y->u.v );   break;
      case 6:
	 d = Vdgg( &x->u.v[3], x->u.v, &y->u.v[3], y->u.v );   break;
      default:
	 execerror("D", "unzulaessige(s) Argument(e)" );
      }
      break;

   default:
      execerror("D", "unzulaessige(s) Argument(e)" );
   }

   r->type = 1;
   r->u.s  = d;
}

void Inhalt(r,v1,v2,v3,v4)
Value *r, *v1, *v2, *v3, *v4;
{
   SKALAR d;

   switch ( r->type )
   {
   case 2:
      if ( v1->type == 3 && v1->type == v2->type )
	 d = Vdist( v1->u.v, v2->u.v );
      else
	 execerror( "I", "unzulaessige Typkombination" );

      break;

   case 3:
      if ( v1->type == 3 && v1->type == v2->type && v2->type == v3->type )
      {
	 SKALAR vh[3];

	 d = fabs( Vnor( 0, v1->u.v, v3->u.v, v2->u.v, vh )) * 0.5;
      }
      else
	 execerror( "I", "unzulaessige Typkombination" );

      break;

   case 4:
      if ( v1->type == 3 && v1->type == v2->type && v2->type == v3->type && v3->type == v4->type )
      {
	 SKALAR vd1[3], vd2[3], vd3[3];

	 Vdif( v1->u.v, v4->u.v, vd1 );
	 Vdif( v2->u.v, v4->u.v, vd2 );
	 Vdif( v3->u.v, v4->u.v, vd3 );
	 d = fabs( Vspat( vd1, vd2, vd3 )) / 6.0;
      }
      else
	 execerror( "I", "unzulaessige Typkombination" );

      break;

   default:
      execerror( "I", "unzulaessige Parameteranzahl" );
   }

   r->type = 1;
   r->u.s  = d;
}

void Winkel(r,x,y)
Value *r, *x, *y;
{
   SKALAR w, v1[3], v2[3];

   if ( r->type > 2 )
      execerror( "W", "unzulaessige Parameteranzahl" );

   if ( r->type == 1 )
      if ( x->type == 5 )
	 w = 2.0 * acos(x->u.v[0]);
      else
	 execerror("W", "unzulaessiges Argument" );
   else
      switch ( x->type )
      {
      case 3:
	 switch ( y->type )
	 {
	 case 3:
	 case 6:
	    w = Vwin( x->u.v, y->u.v );		   break;
	 case 4:
	    w = Vwpe( x->u.v, y->u.v );		   break;
	 default:
	    execerror("W", "unzulaessige(s) Argument(e)" );
	 }
	 break;

      case 4:
	 switch ( y->type )
	 {
	 case 3:
	    w = Vwpe( y->u.v, x->u.v );		   break;
	 case 4:
	 case 6:
	    w = Vwin( y->u.v, x->u.v );		   break;
	 default:
	    execerror("W", "unzulaessige(s) Argument(e)" );
	 }
	 break;

      case 6:
	 switch ( y->type )
	 {
	 case 3:
	 case 4:
	 case 6:
	    w = Vwin( x->u.v, y->u.v );		   break;
	 default:
	    execerror("W", "unzulaessige(s) Argument(e)" );
	 }
	 break;

      default:
	 execerror("W", "unzulaessige(s) Argument(e)" );
      }

   r->type = 1;
   r->u.s  = rng(w);
}

void L_punkt(r,x,y)
Value *r, *x, *y;
{
   SKALAR v, vh[3];

   switch ( x->type )
   {
   case 3:
      switch ( y->type )
      {
      case 4:
	 Vlppe( x->u.v, y->u.v, r->u.v );	    break;
      case 6:
	 Vlppg( x->u.v, &y->u.v[3], y->u.v, r->u.v );    break;
      default:
	 execerror("lp", "unzulaessige(s) Argument(e)" );
      }
      break;

   case 4:
      if ( y->type == 3 )
	 Vlppe( y->u.v, x->u.v, r->u.v );
      else
	 execerror("lp", "unzulaessige(s) Argument(e)" );

      break;

   case 6:
      switch ( y->type )
      {
      case 3:
	 Vlppg( y->u.v, &x->u.v[3], x->u.v, r->u.v );	    break;
      case 6:
	 Vlpgg( &x->u.v[3], x->u.v, &y->u.v[3], y->u.v, r->u.v );    break;
      default:
	 execerror("lp", "unzulaessige(s) Argument(e)" );
      }
      break;

   default:
      execerror("lp", "unzulaessige(s) Argument(e)" );
   }

   r->type = 3;
}

void L_richtung(r,x,y)
Value *r, *x, *y;
{
   short  i;
   SKALAR v, vh[3];

   if ( r->type == 1 )
   {
      if ( x->type == 3 )
	 Vricht(x->u.v,r->u.v);
      else if ( x->type == 4 )
	 for ( i = 0; i < 3; i++ )
	    r->u.v[i] = x->u.v[i];
      else if ( x->type == 5 )
	 Vricht(&x->u.v[1],r->u.v);
      else if ( x->type == 6 )
	 for ( i = 0; i < 3; i++ )
	    r->u.v[i] = x->u.v[i];
      else
	 execerror("rp", "unzulaessige(s) Argument(e)" );
   }
   else if ( r->type == 2 )
      switch ( x->type )
      {
      case 3:
	 switch ( y->type )
	 {
	 case 3:
	    Vlrpp( x->u.v, y->u.v, r->u.v );	    break;
	 case 4:
	    Vlrpe( x->u.v, y->u.v, r->u.v );	    break;
	 case 6:
	    Vlrpg( x->u.v, &y->u.v[3], y->u.v, r->u.v );	    break;
	 default:
	    execerror("rp", "unzulaessige(s) Argument(e)" );
	 }
	 break;

      case 4:
	 if ( y->type == 3 )
	    Vlrpe( y->u.v, x->u.v, r->u.v );
	 else
	    execerror("rp", "unzulaessige(s) Argument(e)" );

	 break;

      case 6:
	 switch ( y->type )
	 {
	 case 3:
	    Vlrpg( y->u.v, &x->u.v[3], x->u.v, r->u.v );	    break;
	 case 6:
	    Vlrgg( &x->u.v[3], x->u.v, &y->u.v[3], y->u.v, r->u.v );    break;
	 default:
	    execerror("rp", "unzulaessige(s) Argument(e)" );
	 }
	 break;

      default:
	 execerror("rp", "unzulaessige(s) Argument(e)" );
      }
   else
      execerror( "rp", "unzulaessige Parameteranzahl" );

   r->type = 3;
}

void L_gerade(r,x,y)
Value *r, *x, *y;
{
   SKALAR v, vh[3];

   if ( x->type == 6 && y->type == 4 )
      Vlgge( &x->u.v[3], x->u.v, y->u.v, &r->u.v[3], r->u.v );
   else if ( x->type == 4 && y->type == 6 )
      Vlgge( &y->u.v[3], y->u.v, x->u.v, &r->u.v[3], r->u.v );
   else
      execerror("lg", "unzulaessige(s) Argument(e)" );

   r->type = 6;
}

void P_mitte(r,v1,v2,v3,v4)
Value *r, *v1, *v2, *v3, *v4;
{
   short n, i;

   n = r->type;
   assval(r,3,v1);

   switch ( n )
   {
   case 4:
      if ( v3->type != v4->type )
	 execerror( "mp", "unzulaessige Typkombination" );

      Vadd(r->u.v,v4->u.v,r->u.v);
   case 3:
      if ( v2->type != v3->type )
	 execerror( "mp", "unzulaessige Typkombination" );

      Vadd(r->u.v,v3->u.v,r->u.v);
   case 2:
      if ( v2->type != v1->type || v1->type != 3 )
	 execerror( "mp", "unzulaessige Typkombination" );

      Vadd(r->u.v,v2->u.v,r->u.v);
      break;

   default:
      execerror( "mp", "unzulaessige Parameteranzahl" );
   }

   for ( i = 0; i < 3; i++ )
      r->u.v[i] /= n;
}

void P_schnitt(r,x,y,z)
Value *r, *x, *y, *z;
{
   if ( r->type == 2 )
   {
      switch ( x->type )
      {
      case 4:
	 if ( y->type == 6 )
	 {
	    if ( !Vspge( &y->u.v[3], y->u.v, x->u.v, x->u.v[3], r->u.v ))
	       execerror("sp", "Schnitt ist leer");
	 }
	 else
	    execerror("sp", "unzulaessige(s) Argument(e)" );

	 break;

      case 6:
	 switch ( y->type )
	 {
	 case 4:
	    if ( !Vspge( &x->u.v[3], x->u.v, y->u.v, y->u.v[3], r->u.v ))
	       execerror("sp", "Schnitt ist leer");

	    break;
	 case 6:
	    if ( !Vspgg( &x->u.v[3], x->u.v, &y->u.v[3], y->u.v, r->u.v ))
	       execerror("sp", "Schnitt ist leer");

	    break;
	 default:
	    execerror("sp", "unzulaessige(s) Argument(e)" );
	 }
	 break;

      default:
	 execerror("sp", "unzulaessige(s) Argument(e)" );
      }
   }
   else if ( r->type == 3 )
   {
      if ( x->type == 4 && y->type == x->type && z->type == y->type )
      {
	 short i;
	 SKALAR A[3][4];

	 for ( i = 0; i < 4; i++ )
	 {
	    A[0][i] = x->u.v[i];
	    A[1][i] = y->u.v[i];
	    A[2][i] = z->u.v[i];
	 }

	 if ( !Vlgs3(A,r->u.v))
	    execerror("sp", "Schnitt ist leer");
      }
      else
	 execerror("sp", "unzulaessige(s) Argument(e)" );
   }
   else
      execerror( "sp", "unzulaessige Parameteranzahl" );

   r->type = 3;
}

void G_schnitt(r,x,y)
Value *r, *x, *y;
{
   if ( x->type == 4 && y->type == x->type )
   {
      short i;
      SKALAR A[3][4];

      Vprod(x->u.v,y->u.v,r->u.v);	  /* Rg := Re1 x Re2  */

      if ( Vabs(r->u.v) < EPSD )
	 execerror("sg", "Ebenen sind parallel" );

      r->u.v[3] = .0;

      for ( i = 0; i < 4; i++ )
      {
	 A[0][i] = x->u.v[i];
	 A[1][i] = y->u.v[i];
	 A[2][i] = r->u.v[i];
      }

      if ( !Vlgs3(A,&r->u.v[3]))
	 execerror("sg", "Schnitt ist leer");

      r->type = 6;
      G_norm(r);
   }
   else
      execerror("sg", "unzulaessige(s) Argument(e)" );
}

void G_verbndg(r,x,y)
Value *r, *x, *y;
{
   if ( x->type == 3 && y->type == x->type )
   {
      short i;

      for ( i = 0; i < 3; i++ )
	 r->u.v[i+3] = x->u.v[i];

      Vdif(y->u.v,x->u.v,r->u.v);
      r->type = 6;
      G_norm(r);
   }
   else
      execerror("vg", "unzulaessige(s) Argument(e)" );
}

void E_verbndg(r,x,y,z)
Value *r, *x, *y, *z;
{
   if ( r->type == 2 )
   {
      switch ( x->type )
      {
      case 3:
	 if ( y->type == 6 )
	    Vvepg( x->u.v, &y->u.v[3], y->u.v, r->u.v );
	 else
	    execerror("ve", "unzulaessige(s) Argument(e)" );

	 break;

      case 6:
	 if ( y->type == 6 )
	    Vvegg( &x->u.v[3], x->u.v, &y->u.v[3], y->u.v, r->u.v );
	 else if ( y->type == 3 )
	    Vvepg( y->u.v, &x->u.v[3], x->u.v, r->u.v );
	 else
	    execerror("ve", "unzulaessige(s) Argument(e)" );

	 break;

      default:
	 execerror("ve", "unzulaessige(s) Argument(e)" );
      }
   }
   else if ( r->type == 3 )
   {
      if ( x->type == 3 && y->type == x->type && z->type == y->type )
	 Vve( x->u.v, y->u.v, z->u.v, r->u.v );
      else
	 execerror("ve", "unzulaessige(s) Argument(e)" );
   }
   else
      execerror( "ve", "unzulaessige Parameteranzahl" );

   r->type = 4;
}

void P_in(r,x,y,z,w)
Value *r, *x, *y, *z, *w;
{
   short t, i;
   SKALAR  A[3][4];

   if (( t = r->type ) == 3 )
   {
      SKALAR  v1[3], v2[3], v3[3], v4[3];

      if ( Vnor(0,x->u.v,y->u.v,z->u.v,A) < EPSD )
	 execerror( "ip", "Vektoren sind kollinear" );

      Vve(x->u.v,y->u.v,z->u.v,A[0]);
      Vdif(y->u.v,x->u.v,v1);
      Vdif(z->u.v,x->u.v,v2);
      Vricht(v1,v1);
      Vricht(v2,v2);
      Vadd(x->u.v,A[0],v3);
      Vadd(x->u.v,v1,v4);
      Vadd(v2,v4,v4);
      Vve(x->u.v,v3,v4,A[1]);
      Vdif(z->u.v,y->u.v,v2);
      Vricht(v2,v2);
      Vadd(y->u.v,A[0],v3);
      Vadd(y->u.v,v2,v4);
      Vdif(v4,v1,v4);
      Vve(y->u.v,v3,v4,A[2]);
   }
   else if ( t == 4 )
   {
      SKALAR  vd1[3][3], vd2[3][3], vd3[3][3];

      Vdif(w->u.v,x->u.v,vd3[0]);
      Vdif(w->u.v,y->u.v,vd3[1]);
      Vdif(w->u.v,z->u.v,vd3[2]);

      if ( Vspat(vd3[0],vd3[1],vd3[2]) < EPSD )
	 execerror( "ip", "Vektoren sind komplanar" );

      Vdif(y->u.v,x->u.v,vd1[0]);
      Vdif(z->u.v,y->u.v,vd1[1]);
      Vdif(x->u.v,z->u.v,vd1[2]);
      Vdif(z->u.v,x->u.v,vd2[0]);
      Vdif(x->u.v,y->u.v,vd2[1]);
      Vdif(y->u.v,z->u.v,vd2[2]);

      for ( i = 0; i < 3; i++ )
      {
	 SKALAR vh[3];

	 Vprod(vd1[i],vd2[i],vh);
	 Vricht(vh,A[i]);
	 Vprod(vd1[i],vd3[i],vh);
	 Vricht(vh,vh);
	 Vadd(A[i],vh,vh);
	 Vricht(vh,A[i]);
      }

      A[0][3] = Vskal(A[0],x->u.v);
      A[1][3] = Vskal(A[1],y->u.v);
      A[2][3] = Vskal(A[2],z->u.v);
   }
   else
      execerror("ip", "unzulaessige Parameteranzahl");

   if ( !Vlgs3(A,r->u.v))
      execerror("ip", "nicht existent");

   r->type = 3;
}

void P_um(r,x,y,z,w)
Value *r, *x, *y, *z, *w;
{
   short t;
   SKALAR  A[3][4], v1[3], v2[3], v3[3];

   if (( t = r->type ) == 3 )
   {
      if ( Vnor(0,x->u.v,y->u.v,z->u.v,A) < EPSD )
	 execerror( "ip", "Vektoren sind kollinear" );

      Vve(x->u.v,y->u.v,z->u.v,A[2]);
   }
   else if ( t == 4 )
   {
      Vdif(w->u.v,x->u.v,v1);
      Vdif(w->u.v,y->u.v,v2);
      Vdif(w->u.v,z->u.v,v3);

      if ( Vspat(v1,v2,v3) < EPSD )
	 execerror( "ip", "Vektoren sind komplanar" );

      Vadd(z->u.v,w->u.v,v3);
      Vdif(w->u.v,z->u.v,A[2]);
      A[2][3] = Vskal(A[2],v3) * 0.5;
   }
   else
      execerror("up", "unzulaessige Parameteranzahl");

   Vadd(x->u.v,y->u.v,v1);
   Vadd(y->u.v,z->u.v,v2);
   Vdif(y->u.v,x->u.v,A[0]);
   A[0][3] = Vskal(A[0],v1) * 0.5;
   Vdif(z->u.v,y->u.v,A[1]);
   A[1][3] = Vskal(A[1],v2) * 0.5;

   if ( !Vlgs3(A,r->u.v))
      execerror("up", "nicht existent");

   r->type = 3;
}
